/*
** Module   : ScreenDefinition.java
** Abstract : contains the configuration of a given frame or family of frames
**
** Copyright (c) 2005-2011, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 EVL 20050705   @21680 Creation based on general concept. Necessary
**                           modification to get compiled.
** 002 EVL 20050713   @21692 Redesign according to the new design
**                           approach.
** 003 EVL 20050715   @21705 Changing the location of the widget types
**                           constants.
** 004 EVL 20050718   @21798 Adding the method to add new created config
**                           to the content of the screen definition. This
**                           will be used when the screen definition is
**                           created on the server side.
** 005 EVL 20050720   @21799 Changes to meet the new XML formats. Each
**                           screen definition holds only one frame, root
**                           or child. So all widgets inside this screen
**                           definition will be the child widgets. Also
**                           the ID variable is added to the class to
**                           identify the parent-child relationship
**                           between the frames if exists.
** 006 EVL 20050727   @21859 Change the method getWidgetsInFrame().
**                           The child frame must also be excluded from
**                           the enumeration process.
** 007 EVL 20050827   @22373 Adding out of the range checking for method
**                           to return the config for the given widget ID.
** 008 SIY 20050831   @22545 Formatting and spelling fixes.
** 009 EVL 20051115   @23336 Moving the code to set up widget's frameID to
**                           the addConfig() method body.
** 010 SIY 20060209   @24450 Fixed Javadoc.
** 011 GES 20060802   @28414 Transmission performance improvements by
**                           shifting to be an externalizable. Many
**                           cleanups to meet coding standards.
** 012 SIY 20060824   @28856 Removed references to DialogBox.
** 013 SIY 20071121   @35930 Support for Frame 'need pause' flag 
**                           maintenance. Some refactoring.
**                           Ref: #35928
** 014 SIY 20080715   @39113 Pause flag is changed to more general 
**                               'suppress redraw' flag.
** 015 SIY 20111101          Moved from CHUI package. Minor 
**                           cleanups/refactorings in preparation to GUI client.
** 016 VIG 20140510          Added tab order list and corresponding methods changes.
*/

package com.goldencode.p2j.ui;

import java.util.*;
import java.io.*;

import com.goldencode.p2j.ui.client.*;

/**
 * This is a container for the configuration of a given frame or family of
 * frames which is efficiently serialized to the client from the server.
 */
public class ScreenDefinition
implements Externalizable
{
   /** The widget configuration data. */
   private LinkedHashMap<Integer, ComponentConfig> widgets = 
                                 new LinkedHashMap<Integer, ComponentConfig>();
   
   /** The list of widget IDs in tab order */
   private LinkedList<Integer> tabOrder = new LinkedList<Integer>();

   /** Root frame ID. */
   private int id = 0;
   
   /** Stores last ID added to the list (for auto-generation of IDs). */
   private int last = 0;
   
   /** Suppress redraw on push of screen definition. */
   private boolean suppressRedraw = false; 
   
   /**
    * Default constructor (only used for serialization).
    */
   public ScreenDefinition()
   {
   }
   
   /**
    * The constructor creates a screen definition with the given root frame
    * ID.
    * 
    * @param    id
    *           Identifies the root frame (or the only frame if this is not
    *           going to hold the configuration of a frame family).
    */
   public ScreenDefinition(int id)
   {
      this.id   = id;
      this.last = id;
   }
   
   /**
    * Get the root frame ID.
    * 
    * @return   The root frame ID.
    */
   public int getRootFrameID()
   {
      return id;
   }
   
   /**
    * Gets an array of all widget IDs in a given frame including any child
    * frames.
    * 
    * @param    frameId
    *           ID of the frame for which widget IDs are requested.
    * 
    * @return   The list of all widget IDs for the specified frame.
    */
   public ArrayList<Integer> getWidgetsInFrame(int frameId)
   {
      ArrayList<Integer> wids = new ArrayList<Integer>();
      Iterator<ComponentConfig> iter = getConfigsIterator();
 
      while (iter.hasNext())
      {
         ComponentConfig config = iter.next();
         
         if (config.getFrameId() == frameId && config.getType() != ComponentConfig.FRAME)
         {
            wids.add(new Integer(config.getID()));
         }
      }

      return wids;
   }
   
   /**
    * Get an iterator to enumerate all widget configurations inside this
    * screen definition.
    * 
    * @return   The iterator.
    */
   public Iterator<ComponentConfig> getConfigsIterator()
   {
      return widgets.values().iterator();
   }
   
   /**
    * Returns widgets ids list corresponding with tab order
    */
   public LinkedList<Integer> getTabOrder() 
   {
      return tabOrder;
   }
   
   /**
    * Gets a specific widget configuration instance for the widget with the
    * ID provided.
    * 
    * @param    id
    *           The widget ID.
    * 
    * @return   The requested widget configuration object if exists,
    *           <code>null</code> otherwise.
    */
   public ComponentConfig getConfig(int id)
   {
      return widgets.get(new Integer(id));
   }
   
   /**
    * Add a widget configuration to the current screen definition.
    * 
    * @param    config
    *           New config to be added to the current screen definition.
    */
   public void addConfig(ComponentConfig config)
   {
      addConfigInt(config);
   }

   /**
    * Add a widget configuration (to the current screen definition) and
    * assign that widget the specified ID.
    * 
    * @param    config
    *           New config to be added to the current screen definition.
    * @param    widgetID
    *           Widget ID.
    */
   public void addConfig(ComponentConfig config, int widgetID)
   {
      config.setID(widgetID);
      addConfigInt(config);
   }
   
   /**
    * Replacement for the default object reading method. The latest
    * state is read from the input source.
    * 
    * @param    in
    *           The input source from which fields will be restored.
    *
    * @throws   IOException
    *           In case of I/O errors.
    * @throws   ClassNotFoundException
    *           If payload can't be instantiated.
    */
   @Override
   public void readExternal(ObjectInput in)
   throws IOException,
          ClassNotFoundException
   {
      id             = in.readInt();
      suppressRedraw = in.readBoolean();
      WidgetEntry.arrayToMap((WidgetEntry[]) in.readObject(), widgets);
      
      Integer[] tabOrderArr = (Integer[]) in.readObject();
      if (tabOrderArr != null)
      {
         for (int i = 0; i < tabOrderArr.length; i++)
            tabOrder.add(tabOrderArr[i]);
      }
   }

   /**
    * Replacement for the default object writing method. The latest
    * state is written to the output destination.
    * 
    * @param    out
    *           The output destination to which fields will be saved.
    *
    * @throws   IOException
    *           In case of I/O errors.
    */
   @Override
   public void writeExternal(ObjectOutput out)
   throws IOException
   {
      out.writeInt(id);
      out.writeBoolean(suppressRedraw);
      out.writeObject(WidgetEntry.mapToArray(widgets));
      out.writeObject(tabOrder.toArray(new Integer[tabOrder.size()]));
   }   
   
   /**
    * Calculates the first widget in the tab order for the specified field-group.
    * 
    * @param   fieldGroup
    *          fieldGroup to get the first widget in its tab order.
    * 
    * @return  The ID of the first widget.
    */
   public Integer getFirstTabItem(FieldGroup fieldGroup)
   {
      // TODO add fieldGroup handling
      return tabOrder.getFirst();
   }

   /**
    * Calculates the next widget in the tab order of the specified widget.
    * 
    * @param   fieldGroup
    *          fieldGroup to get the first widget in its tab order.
    * @param   widgetId
    *          The source widget to calculate the next widget from.
    * 
    * @return  The ID of the next widget.
    */
   public Integer getNextTabItem(FieldGroup fieldGroup, Integer widgetId)
   {
      // TODO add fieldGroup handling
      if (!tabOrder.contains(widgetId))
      {
         throw new IllegalArgumentException("widgetId must be valid widget ids in this frame");
      }
      
      int idx = tabOrder.indexOf(widgetId);
      if (idx == -1)
      {
         throw new IllegalStateException("The widget with specified id is not present in tab order");
      }
      
      if (tabOrder.getLast() == widgetId)
      {
         return null;
      }
      
      return tabOrder.get(idx + 1);
   }
   
   /**
    * Assigns the current widget to the tab position before the specified in the same field group.
    *
    * @param   widget
    *          The handle of the widget before whose we need to make new assignment.
    */
   public void moveBeforeTabItem(Integer widgetId, Integer beforeId)
   {
      if (widgetId == null || beforeId == null)
      {
         throw new NullPointerException("widgetId and beforeId must be specified");
      }

      if (!widgets.containsKey(widgetId) || !widgets.containsKey(beforeId))
      {
         throw new IllegalArgumentException(
                  "widgetId and beforeId must be valid widget ids in this frame");
      }

      if (!tabOrder.contains(widgetId) || !tabOrder.contains(beforeId))
      {
         throw new IllegalArgumentException(
                  "widgetId and beforeId must be valid widget ids in this frame");
      }
      
      tabOrder.remove((Object)widgetId);
      
      tabOrder.add(tabOrder.indexOf(beforeId), widgetId);
   }
   
   /**
    * Common worker to add a widget configuration to the screen definition.
    * 
    * @param    config
    *           Widget configuration.
    */
   private void addConfigInt(ComponentConfig config)
   {
      if (config.getFrameId() == -1)
      {
         if (config.getType() != ComponentConfig.FRAME)
         {
            config.setFrameId(id);
         }
      }
      
      int wid = config.getID();
      
      if (wid == -1)
      {
         // auto-generate an ID
         while (widgets.get(new Integer(++last)) != null) { };
         
         config.setID(last);
         wid = last;
      }
      else
      {
         // save the latest widget's ID as the last
         last = wid;
      }
      
      Integer iwid = new Integer(wid);
      
      if (widgets.get(iwid) != null)
      {
         throw new RuntimeException("Duplicate widget ID = " + iwid);
      }
      
      widgets.put(iwid, config);
      
      if(config.getType() != ComponentConfig.FRAME) // TODO delete if
         tabOrder.add(iwid);
   }

   /**
    * Set new value for 'suppress redraw' flag.
    * 
    * @param   suppressRedraw
    *          New value of flag.
    */
   public void setSuppressRedraw(boolean suppressRedraw)
   {
      this.suppressRedraw = suppressRedraw; 
   }

   /**
    * Get current value of 'suppress redraw' flag.
    * 
    * @return  The current value of the flag. 
    */
   public boolean isSuppressRedraw()
   {
      return suppressRedraw; 
   }
}
