/*
** Module   : ScrollPane.java
** Abstract : ScrollPane helper class implementation for CHUI
**
** Copyright (c) 2005-2011, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 EVL 20050203   @19739 Creation based on CHARVA JScrollPane class.
**                           Necessary modification to get compiled.
** 002 EVL 20050307   @20218 Modification of the scrollbar's slider
**                           position calculation. The new algorithm takes
**                           into account the negative coordinate value
**                           for viewport Y position.
** 003 EVL 20050408   @20672 Adding horizontal scroll-bar but not for any
**                           instances of JList. SelectionListBody and
**                           ComboBox are the such a cases for now
** 004 EVL 20050421   @20831 Changed constructor order.
** 005 EVL 20050429   @20944 Adding label creation for the widget.
** 006 EVL 20050503   @21072 Setting up the default one line border.
** 007 EVL 20050522   @21252 Adding conditions to choose the child drawing
**                           If the child is instance of the JList or
**                           parent is the Frame or DialogBox we call the
**                           super.draw(), otherwise we handle clipping in
**                           the FrameHelper and have to call only draw()
**                           method of the child.
** 008 EVL 20050531   @21337 Changing the inner variable names to sync
**                           with the rest of the project and preparing
**                           to the javadoc processing.
** 009 EVL 20050607   @21435 Syntax for comments checking.
** 010 EVL 20050729   @21915 Removing the ugly tabulation from source.
** 011 EVL 20050825   @22273 Adding method to get wrapped component
**                           inside.
** 012 SIY 20050831   @22548 Fixed formatting and spelling.
** 013 SIY 20050926   @22848 Using ColorTable for drawing.
** 014 SIY 20060209   @24451 Cleanups and fixes for new SelectionList.
** 015 SIY 20060209   @24498 Fixed scroll-bar handling.
** 016 SIY 20060222   @24641 Removed references to DialogBox.
** 017 GES 20060223   @24680 Eliminated cursor movement in no-box case.
**                           This caused problems in redirected stream
**                           output.
** 018 SIY 20060518   @26436 Added separate drawing routine for use in 
**                           down frames.
** 019 SIY 20060518   @26438 Fixed misspelled method name.
** 020 SIY 20060524   @26459 Added drawing of the border.
** 021 SIY 20060525   @26536 Fixed drawing of the child content.
** 022 NVS 20060728   @28214 Fixed drawing of the border in redirected
**                           mode.
** 023 SIY 20060824   @28857 Removed reference to DialogBox and getP2JType
**                           method.
** 024 GES 20061012   @30369 Don't draw scrollbars when in redirected
**                           mode.
** 025 NVS 20061221   @31731 Fixed drawing of contained selection lists.
** 026 SIY 20070727   @34716 Do not draw scroll bars if their size is less
**                           than 3 chars.
** 027 SIY 20080214   @37112 Fixed drawing of parts of frame which don't 
**                           fit into visible area.
**                           Ref: #37107
** 028 NVS 20080401   @37780 Added isFocusable() method implementation.
** 029 SIY 20080401   @37795 Do not draw scroll-bars if there are borders.
**                           Fix for #37745.
** 030 SIY 20080505   @38206 Draw scroll-bars for DropDown even there are
**                           no borders.
** 031 NVS 20080828   @39621 setLocation() method now calls super class
**                           as well.
** 032 SIY 20081125   @40691 Removed unnecessary scroll() method.
** 033 NVS 20090213   @41316 Dead code removal.
** 034 SIY 20100619          Switching to own implementation of character UI.
** 035 SIY 20100804          Changes in imports, some refactorings.
** 036 SIY 20100822          Refactored AttributeContainer support.
** 037 SIY 20101122          Reduced class hierarchy. Switching from KeyEvent 
**                           to KeyInput.
** 038 SIY 20110112          Changes related to new color handling scheme.
** 039 SIY 20111202          Refactored and moved from CHUI package.
** 040 VIG 20140515          Added tab order filling.
*/

package com.goldencode.p2j.ui.client;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.border.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.event.listener.*;
import com.goldencode.p2j.ui.client.widget.*;

/**
 * Provides a scrollable view service for frames. The component to be placed
 * inside this class must implement the <code>Scrollable</code> interface.
 */
public abstract class ScrollPane
extends AbstractContainer
implements ScrollListener
{
   /** Border around the pane. */
   private Border border;

   /** View port for the visible part of the pane. */
   private Viewport viewport;

   /**  The component which is wrapped by this ScrollPane. */
   protected Widget child = null;

   /** New label to be used with this container. */
   private String label = null;

   /** Attribute container. */
   private BaseConfig config = null;

   /**
    * Create a ScrollPane that displays the contents of the specified
    * widget.
    * 
    * @param   widget
    *          The widget to be displayed. This widget must implement
    *          the {@link Scrollable} interface.
    */
   public ScrollPane(Widget widget)
   {
      viewport = new Viewport(widget);
      add(viewport);

      if (widget instanceof Scrollable)
         ((Scrollable) widget).addScrollListener(this);
   
      child = widget;
      
      if ("" != null && "".length() > 0)
         label = new String("" + ":");
      
      setBorder(new LineBorder());
      config = new BaseConfig(ComponentConfig.SCROLL_PANE, this);
   }

   /**
    * Get access to border.
    * 
    * @return  border reference.
    */
   public Border border()
   {
      return border;
   }

   /**
    * Provide access to the attributes stored in the widget.
    * 
    * @return  Reference to attribute container.
    */
   @Override
   @SuppressWarnings("unchecked")
   public ComponentConfig config()
   {
      return config;
   }

   /**
    * The main renderer for this container. Draws the child container and
    * vertical/horizontal scroll-bars as necessary.
    */
   @Override
   public abstract void draw();

   /**
    * The secondary renderer for this container. Draws the child container and
    * vertical/horizontal scroll-bars as necessary. Note that this routine
    * assumes that it is invoked from Frame class.
    * 
    * @param   list
    *          Array of components to draw.
    */
   public abstract void draw(Widget[] list);

   /**
    * Getting the wrapped component from the scroll pane. Used for the
    * selection list widget to request the widget handle.
    * 
    * @return  The handle of the component wrapped by the scroll pane.
    */
   public Widget getChild()
   {
      return child;
   }

   /**
    * Checks whether this attribute container can receive focus generally.
    * 
    * @return  <code>true</code> if this container can be focused under
    *          certain conditions. <code>false</code> if, and only if, this
    *          attribute container can never receive focus.    
    */
   @Override
   public boolean isFocusable()
   {
      return false;
   }

   /**
    * Get widget minimum size.
    * 
    * @return  widget minimum size.
    */
   @Override
   public Dimension minimumSize()
   {
      Widget view = viewport().view();

      Dimension size = (view instanceof Scrollable) ? 
                                          ((Scrollable) view).preferredSize()
                                          : view.dimension();

      viewport.setExtent(size.width, size.height);

      if (border != null)
         size = size.enlargedBy(border.insets());

      return size;
   }

   /**
    * This method is invoked when scroll pane is scrolled.
    * 
    * @param   event
    *          Event which triggered this call.
    */
   @Override
   public void scroll(ScrollEvent event)
   {
      ScrollDirection direction = event.direction();

      Point limit = event.limit();

      limit.translate(viewport().location());
      limit.translate(event.scrollable().location());

      Rectangle rect = viewport.bounds();
      Point pos = viewport.view().location();

      if (limit.x < rect.left() || limit.x > rect.right())
      {
         if (ScrollDirection.left(direction))
         {
            pos.x -= limit.x - rect.right();
         }
         else if (ScrollDirection.right(direction))
         {
            pos.x += rect.left() - limit.x;
         }
      }
      
      if (limit.y < rect.top() && ScrollDirection.down(direction))
      {
         pos.y += rect.top() - limit.y;
      }
      
      if (limit.y > rect.bottom() && ScrollDirection.up(direction))
      {
         pos.y -= limit.y - rect.bottom();
      }

      viewport.view().setLocation(pos);
   }

   /**
    * Set pane border.
    *  
    * @param   border
    *          New pane border.
    */
   public void setBorder(Border border)
   {
      this.border = border;
      viewport.setLocation(border.insets().asPoint());
   }

   /**
    * Setting the location of this component.
    * 
    * @param   x
    *          The new x(column) location relative to the parent container.
    * @param   y
    *          The new y(row) location relative to the parent container.
    */
   @Override
   public void setLocation(int x, int y)
   {
      super.setLocation(x, y);
   }

   /**
    * Setting the location of this component.
    * 
    * @param   point
    *          The Point object to be used as the new location of the
    *          container.
    */
   @Override
   public void setLocation(Point point)
   {
      if (label != null)
         point.x += label.length() + 1;
      setLocation(point.x, point.y);
   }

   /**
    * Set container size.
    * 
    * @param   size
    *          Size to set.
    */
   @Override
   public void setSize(Dimension size)
   {
      setSize(size.width, size.height);
   }

   /**
    * Set container size.
    * 
    * @param   width
    *          New width.
    * @param   height
    *          New height.
    */
   @Override
   public void setSize(int width, int height)
   {
      super.setSize(width, height);

      Dimension size = new Dimension(width, height);

      if (border != null)
         size = size.reducedBy(border.insets());

      viewport.setExtent(size.width, size.height);
   }

   /**
    * Get pane viewport.
    * 
    * @return  pane viewport.
    */
   public Viewport viewport()
   {
      return viewport;
   }

   /**
    * Calculate horizontal thumbnail position.
    * 
    * @param   childSize
    *          The size of the child pane.
    * @param   extentSize
    *          The size of the viewport.
    * @param   viewPosition
    *          The current view position.
    * 
    * @return  The horizontal thumbnail position.
    */
   protected int calcHscroll(Dimension childSize, Dimension extentSize, Point viewPosition)
   {
      int offset = (extentSize.width - 2) * (-viewPosition.x)
                   / (childSize.width - extentSize.width);
      if (offset == 0)
         offset = 1;
      return offset;
   }

   /**
    * Calculate vertical thumbnail position.
    * 
    * @param   childSize
    *          The size of the child pane.
    * @param   extentSize
    *          The size of the viewport.
    * @param   viewPosition
    *          The current view position.
    *          
    * @return  The vertical thumbnail position.
    */
   protected int calcVscroll(Dimension childSize, Dimension extentSize, Point viewPosition)
   {
      int offset = 1;
   
      if (childSize.height > extentSize.height)
         offset = (extentSize.height - 2) * (-viewPosition.y)
                  / (childSize.height - extentSize.height);
      else if (child instanceof ScrollableList)
         offset = ((ScrollableList) child).getCurrentRow() + 1;
   
      if (offset <= 0)
         offset = 1;
      if (offset > extentSize.height - 2)
         offset = extentSize.height - 2;
      return offset;
   }

   /**
    * 
    */
   protected void drawDirect()
   {
      super.draw();
   }
}
